@@ -10,7 +10,6 @@
 #include "mozilla/dom/Document.h"
 #include "nsGkAtoms.h"
 #include "nsMenuPopupFrame.h"
-#include "nsIServiceManager.h"
 #include "nsIDragService.h"
 #include "nsIDragSession.h"
 #ifdef MOZ_XUL
@@ -22,12 +21,14 @@
 #  include "nsXULPopupManager.h"
 #endif
 #include "nsIPopupContainer.h"
+#include "nsServiceManagerUtils.h"
 #include "nsTreeColumns.h"
 #include "nsContentUtils.h"
 #include "mozilla/ErrorResult.h"
 #include "mozilla/LookAndFeel.h"
 #include "mozilla/Preferences.h"
 #include "mozilla/PresShell.h"
+#include "mozilla/ScopeExit.h"
 #include "mozilla/dom/Element.h"
 #include "mozilla/dom/Event.h"  // for Event
 #include "mozilla/dom/MouseEvent.h"
@@ -54,7 +55,7 @@ nsXULTooltipListener::nsXULTooltipListener()
       mLastTreeRow(-1)
 #endif
 {
-  // FIXME(emilio): This can be faster, this should use BoolVarCache.
+  // FIXME(emilio): This can be faster, this should use static prefs.
   //
   // register the callback so we get notified of updates
   Preferences::RegisterCallback(ToolbarTipsPrefChanged,
@@ -112,9 +113,15 @@ void nsXULTooltipListener::MouseOut(Event* aEvent) {
     if (pm) {
       nsCOMPtr<nsINode> tooltipNode =
           pm->GetLastTriggerTooltipNode(currentTooltip->GetComposedDoc());
-      if (tooltipNode == targetNode) {
-        // if the target node is the current tooltip target node, the mouse
-        // left the node the tooltip appeared on, so close the tooltip.
+
+      // If the target node is the current tooltip target node, the mouse
+      // left the node the tooltip appeared on, so close the tooltip. However,
+      // don't do this if the mouse moved onto the tooltip in case the
+      // tooltip appears positioned near the mouse.
+      nsCOMPtr<EventTarget> relatedTarget =
+          aEvent->AsMouseEvent()->GetRelatedTarget();
+      nsCOMPtr<nsIContent> relatedContent = do_QueryInterface(relatedTarget);
+      if (tooltipNode == targetNode && relatedContent != currentTooltip) {
         HideTooltip();
         // reset special tree tracking
         if (mIsSourceTree) {
@@ -202,7 +209,7 @@ void nsXULTooltipListener::MouseMove(Event* aEvent) {
     if (mTargetNode) {
       nsresult rv = NS_NewTimerWithFuncCallback(
           getter_AddRefs(mTooltipTimer), sTooltipCallback, this,
-          LookAndFeel::GetInt(LookAndFeel::eIntID_TooltipDelay, 500),
+          LookAndFeel::GetInt(LookAndFeel::IntID::TooltipDelay, 500),
           nsITimer::TYPE_ONE_SHOT, "sTooltipCallback",
           sourceContent->OwnerDoc()->EventTargetFor(TaskCategory::Other));
       if (NS_FAILED(rv)) {
@@ -227,7 +234,7 @@ NS_IMETHODIMP
 nsXULTooltipListener::HandleEvent(Event* aEvent) {
   nsAutoString type;
   aEvent->GetType(type);
-  if (type.EqualsLiteral("DOMMouseScroll") || type.EqualsLiteral("mousedown") ||
+  if (type.EqualsLiteral("wheel") || type.EqualsLiteral("mousedown") ||
       type.EqualsLiteral("mouseup") || type.EqualsLiteral("dragstart")) {
     HideTooltip();
     return NS_OK;
@@ -293,16 +300,11 @@ void nsXULTooltipListener::AddTooltipSupport(nsIContent* aNode) {
   MOZ_ASSERT(aNode);
   MOZ_ASSERT(this == sInstance);
 
-  aNode->AddSystemEventListener(NS_LITERAL_STRING("mouseout"), this, false,
-                                false);
-  aNode->AddSystemEventListener(NS_LITERAL_STRING("mousemove"), this, false,
-                                false);
-  aNode->AddSystemEventListener(NS_LITERAL_STRING("mousedown"), this, false,
-                                false);
-  aNode->AddSystemEventListener(NS_LITERAL_STRING("mouseup"), this, false,
-                                false);
-  aNode->AddSystemEventListener(NS_LITERAL_STRING("dragstart"), this, true,
-                                false);
+  aNode->AddSystemEventListener(u"mouseout"_ns, this, false, false);
+  aNode->AddSystemEventListener(u"mousemove"_ns, this, false, false);
+  aNode->AddSystemEventListener(u"mousedown"_ns, this, false, false);
+  aNode->AddSystemEventListener(u"mouseup"_ns, this, false, false);
+  aNode->AddSystemEventListener(u"dragstart"_ns, this, true, false);
 }
 
 void nsXULTooltipListener::RemoveTooltipSupport(nsIContent* aNode) {
@@ -312,11 +314,11 @@ void nsXULTooltipListener::RemoveTooltipSupport(nsIContent* aNode) {
   // The last reference to us can go after some of these calls.
   RefPtr<nsXULTooltipListener> instance = this;
 
-  aNode->RemoveSystemEventListener(NS_LITERAL_STRING("mouseout"), this, false);
-  aNode->RemoveSystemEventListener(NS_LITERAL_STRING("mousemove"), this, false);
-  aNode->RemoveSystemEventListener(NS_LITERAL_STRING("mousedown"), this, false);
-  aNode->RemoveSystemEventListener(NS_LITERAL_STRING("mouseup"), this, false);
-  aNode->RemoveSystemEventListener(NS_LITERAL_STRING("dragstart"), this, true);
+  aNode->RemoveSystemEventListener(u"mouseout"_ns, this, false);
+  aNode->RemoveSystemEventListener(u"mousemove"_ns, this, false);
+  aNode->RemoveSystemEventListener(u"mousedown"_ns, this, false);
+  aNode->RemoveSystemEventListener(u"mouseup"_ns, this, false);
+  aNode->RemoveSystemEventListener(u"dragstart"_ns, this, true);
 }
 
 #ifdef MOZ_XUL
@@ -348,12 +350,7 @@ void nsXULTooltipListener::CheckTreeBodyMove(MouseEvent* aMouseEvent) {
     // determine if we are going to need a titletip
     // XXX check the disabletitletips attribute on the tree content
     mNeedTitletip = false;
-    int16_t colType = -1;
-    if (col) {
-      colType = col->Type();
-    }
-    if (row >= 0 && cellInfo.mChildElt.EqualsLiteral("text") &&
-        colType != TreeColumn_Binding::TYPE_PASSWORD) {
+    if (row >= 0 && cellInfo.mChildElt.EqualsLiteral("text")) {
       mNeedTitletip = tree->IsCellCropped(row, col, rv);
     }
 
@@ -399,10 +396,10 @@ nsresult nsXULTooltipListener::ShowTooltip() {
 
       // listen for popuphidden on the tooltip node, so that we can
       // be sure DestroyPopup is called even if someone else closes the tooltip
-      currentTooltip->AddSystemEventListener(NS_LITERAL_STRING("popuphiding"),
-                                             this, false, false);
+      currentTooltip->AddSystemEventListener(u"popuphiding"_ns, this, false,
+                                             false);
 
-      // listen for mousedown, mouseup, keydown, and DOMMouseScroll events at
+      // listen for mousedown, mouseup, keydown, and mouse events at
       // document level
       Document* doc = sourceNode->GetComposedDoc();
       if (doc) {
@@ -411,13 +408,12 @@ nsresult nsXULTooltipListener::ShowTooltip() {
         // applications.  If we don't specify the aWantsUntrusted of
         // AddSystemEventListener(), the event target sets it to TRUE if the
         // target is in content.
-        doc->AddSystemEventListener(NS_LITERAL_STRING("DOMMouseScroll"), this,
-                                    true);
-        doc->AddSystemEventListener(NS_LITERAL_STRING("mousedown"), this, true);
-        doc->AddSystemEventListener(NS_LITERAL_STRING("mouseup"), this, true);
+        doc->AddSystemEventListener(u"wheel"_ns, this, true);
+        doc->AddSystemEventListener(u"mousedown"_ns, this, true);
+        doc->AddSystemEventListener(u"mouseup"_ns, this, true);
 #ifndef XP_WIN
         // On Windows, key events don't close tooltips.
-        doc->AddSystemEventListener(NS_LITERAL_STRING("keydown"), this, true);
+        doc->AddSystemEventListener(u"keydown"_ns, this, true);
 #endif
       }
       mSourceNode = nullptr;
@@ -444,8 +440,10 @@ static void SetTitletipLabel(XULTreeElement* aTree, Element* aTooltip,
 #endif
 
 void nsXULTooltipListener::LaunchTooltip() {
-  nsCOMPtr<Element> currentTooltip = do_QueryReferent(mCurrentTooltip);
-  if (!currentTooltip) return;
+  RefPtr<Element> currentTooltip = do_QueryReferent(mCurrentTooltip);
+  if (!currentTooltip) {
+    return;
+  }
 
 #ifdef MOZ_XUL
   if (mIsSourceTree && mNeedTitletip) {
@@ -456,25 +454,35 @@ void nsXULTooltipListener::LaunchTooltip() {
       // Because of mutation events, currentTooltip can be null.
       return;
     }
-    currentTooltip->SetAttr(kNameSpaceID_None, nsGkAtoms::titletip,
-                            NS_LITERAL_STRING("true"), true);
+    currentTooltip->SetAttr(kNameSpaceID_None, nsGkAtoms::titletip, u"true"_ns,
+                            true);
   } else {
     currentTooltip->UnsetAttr(kNameSpaceID_None, nsGkAtoms::titletip, true);
   }
+
   if (!(currentTooltip = do_QueryReferent(mCurrentTooltip))) {
     // Because of mutation events, currentTooltip can be null.
     return;
   }
 
   nsXULPopupManager* pm = nsXULPopupManager::GetInstance();
-  if (pm) {
-    nsCOMPtr<nsIContent> target = do_QueryReferent(mTargetNode);
-    pm->ShowTooltipAtScreen(currentTooltip, target, mMouseScreenX,
-                            mMouseScreenY);
-
-    // Clear the current tooltip if the popup was not opened successfully.
-    if (!pm->IsPopupOpen(currentTooltip)) mCurrentTooltip = nullptr;
+  if (!pm) {
+    return;
   }
+
+  auto cleanup = MakeScopeExit([&] {
+    // Clear the current tooltip if the popup was not opened successfully.
+    if (!pm->IsPopupOpen(currentTooltip)) {
+      mCurrentTooltip = nullptr;
+    }
+  });
+
+  RefPtr<Element> target = do_QueryReferent(mTargetNode);
+  if (!target) {
+    return;
+  }
+
+  pm->ShowTooltipAtScreen(currentTooltip, target, mMouseScreenX, mMouseScreenY);
 #endif
 }
 
@@ -631,19 +639,16 @@ nsresult nsXULTooltipListener::DestroyTooltip() {
     nsCOMPtr<Document> doc = currentTooltip->GetComposedDoc();
     if (doc) {
       // remove the mousedown and keydown listener from document
-      doc->RemoveSystemEventListener(NS_LITERAL_STRING("DOMMouseScroll"), this,
-                                     true);
-      doc->RemoveSystemEventListener(NS_LITERAL_STRING("mousedown"), this,
-                                     true);
-      doc->RemoveSystemEventListener(NS_LITERAL_STRING("mouseup"), this, true);
+      doc->RemoveSystemEventListener(u"wheel"_ns, this, true);
+      doc->RemoveSystemEventListener(u"mousedown"_ns, this, true);
+      doc->RemoveSystemEventListener(u"mouseup"_ns, this, true);
 #ifndef XP_WIN
-      doc->RemoveSystemEventListener(NS_LITERAL_STRING("keydown"), this, true);
+      doc->RemoveSystemEventListener(u"keydown"_ns, this, true);
 #endif
     }
 
     // remove the popuphidden listener from tooltip
-    currentTooltip->RemoveSystemEventListener(NS_LITERAL_STRING("popuphiding"),
-                                              this, false);
+    currentTooltip->RemoveSystemEventListener(u"popuphiding"_ns, this, false);
   }
 
   // kill any ongoing timers
